searchData={"items":[{"type":"module","title":"Agens","doc":"Agens is used to create multi-agent workflows with language models.\n\nIt is made up of the following core entities:\n\n- `Agens.Serving` - used to interact with language models\n- `Agens.Agent` - used to interact with servings in a specialized manner\n- `Agens.Job` - used to define multi-agent workflows","ref":"Agens.html"},{"type":"function","title":"Agens.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Agens.html#child_spec/1"},{"type":"module","title":"Agens.Message","doc":"The Message struct defines the details of a message passed between Agents, Jobs and Servings.","ref":"Agens.Message.html"},{"type":"module","title":"Fields - Agens.Message","doc":"* `:parent_pid` - The process identifier of the parent/caller process.\n  * `:input` - The input string for the message.\n  * `:prompt` - The prompt string or `Agens.Agent.Prompt` struct for the message.\n  * `:result` - The result string for the message.\n  * `:agent_name` - The name of the `Agens.Agent`.\n  * `:serving_name` - The name of the `Agens.Serving`.\n  * `:job_name` - The name of the `Agens.Job`.\n  * `:job_description` - The description of the `Agens.Job` to be added to the LM prompt.\n  * `:step_index` - The index of the `Agens.Job.Step`.\n  * `:step_objective` - The objective of the `Agens.Job.Step` to be added to the LM prompt.","ref":"Agens.Message.html#module-fields"},{"type":"function","title":"Agens.Message.send/1","doc":"Sends an `Agens.Message` to an `Agens.Agent`","ref":"Agens.Message.html#send/1"},{"type":"type","title":"Agens.Message.t/0","doc":"","ref":"Agens.Message.html#t:t/0"},{"type":"module","title":"Agens.Supervisor","doc":"The Supervisor module for the Agens application.\n\n`Agens.Supervisor` starts a `DynamicSupervisor` for managing `Agens.Agent`, `Agens.Serving`, and `Agens.Job` processes. It also starts a `Registry` for keeping track of these processes.\n\nThe Registry module can be overriden by your application config:\n\n```\nconfig :agens, registry: MyApp.Registry\n```\n\nIn order to use `Agens` simply add `Agens.Supervisor` to your application supervision tree:\n\n```\nSupervisor.start_link(\n  [\n    {Agens.Supervisor, name: Agens.Supervisor}\n  ],\n  strategy: :one_for_one\n)\n```","ref":"Agens.Supervisor.html"},{"type":"function","title":"Agens.Supervisor.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Agens.Supervisor.html#child_spec/1"},{"type":"module","title":"Agens.Agent","doc":"The Agent module provides functions for starting, stopping and running Agents.\n\n`Agens.Agent` is the the primary entity powering `Agens`. It uses `Agens.Serving` to interact with language models through `Nx.Serving`, or with language model APIs through a `GenServer`.\n\nAgents can have detailed identities to further refine LM outputs, and are used together in multi-agent workflows via the `Agens.Job` module.\n\nAgent capabilities can be expanded even further with `Agens.Tool` modules, which are designed to handle LM functional calling. In future releases, Agents will also have access to RAG generations via knowledge base features.","ref":"Agens.Agent.html"},{"type":"function","title":"Agens.Agent.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Agens.Agent.html#child_spec/1"},{"type":"function","title":"Agens.Agent.start/1","doc":"Starts one or more `Agens.Agent` processes","ref":"Agens.Agent.html#start/1"},{"type":"function","title":"Agens.Agent.stop/1","doc":"Stops an `Agens.Agent` process","ref":"Agens.Agent.html#stop/1"},{"type":"module","title":"Agens.Agent.Config","doc":"The Config struct represents the configuration for an Agent process.","ref":"Agens.Agent.Config.html"},{"type":"module","title":"Fields - Agens.Agent.Config","doc":"- `:name` - The name of the Agent process.\n- `:serving` - The serving module or `Nx.Serving` struct for the Agent.\n- `:knowledge` - The knowledge base or data source of the Agent. Default is nil. (Coming soon)\n- `:prompt` - The string or `Agens.Agent.Prompt` struct defining the Agent. Default is nil.\n- `:tool` - The module implementing the `Agens.Tool` behaviour for the Agent. Default is nil.","ref":"Agens.Agent.Config.html#module-fields"},{"type":"type","title":"Agens.Agent.Config.t/0","doc":"","ref":"Agens.Agent.Config.html#t:t/0"},{"type":"module","title":"Agens.Agent.Prompt","doc":"The Prompt struct represents an advanced prompt for an Agent process.\n\nAll fields are optional and will only be included in the final prompt if they are not nil.","ref":"Agens.Agent.Prompt.html"},{"type":"module","title":"Fields - Agens.Agent.Prompt","doc":"- `:identity` - a string representing the purpose and capabilities of the agent\n- `:context` - a string representing the goal or purpose of the agent's actions\n- `:constraints` - a string listing any constraints or limitations on the agent's actions\n- `:examples` - a list of example inputs and outputs for the agent\n- `:reflection` - a string representing any additional considerations or reflection the agent should make before returning results","ref":"Agens.Agent.Prompt.html#module-fields"},{"type":"type","title":"Agens.Agent.Prompt.t/0","doc":"","ref":"Agens.Agent.Prompt.html#t:t/0"},{"type":"module","title":"Agens.Job","doc":"A Job defines a multi-agent workflow through a sequence of steps.\n\nAn `Agens.Job` is mainly a sequence of steps, defined with the `Agens.Job.Step` struct, used to create advanced multi-agent workflows.\n\nConditions can be used in order to route to different steps based on a result, or can be used to end the Job.","ref":"Agens.Job.html"},{"type":"module","title":"Events - Agens.Job","doc":"Agens emits several events that can be handled by the caller using `handle_info/3` for purposes such as UI updates, pubsub, logging, persistence and other side effects.\n\n#### Job\n```\n{:job_started, job.name}\n```\n\nEmitted when a job has started.\n\n```\n{:job_ended, job.name, :completed | {:error, error}}\n```\n\nEmitted when a job has ended, either due to completion or an error.\n\n#### Step\n```\n{:step_started, {job.name, step_index}, message.input}\n```\n\nEmitted when a step has started. Includes the input data provided to the step, whether from the user or a previous step.\n\n```\n{:step_result, {job.name, step_index}, message.result}\n```\n\nEmitted when a result has been returned from the Serving. Includes the Serving result, which will be passed to the Tool (if applicable), conditions (if applicable), or the next step of the job.\n\n#### Tool\nThe following events are emitted only if the Agent has a Tool specified in `Agens.Agent.Config`:\n\n```\n{:tool_started, {job.name, step_index}, message.result}\n```\n\nEmitted when a Tool is about to be called. `message.result` here is the Serving result, which will be overriden by the value returned from the Tool prior to final output.\n\n```\n{:tool_raw, {job.name, step_index}, message.raw}\n```\n\nEmitted after completing the Tool function call. It provides the raw result of the Tool before any post-processing.\n\n```\n{:tool_result, {job.name, step_index}, message.result}\n```\n\nEmitted after post-processing of the raw Tool result. This is the final result of the Tool, which will be passed to conditions or the next step of the job.","ref":"Agens.Job.html#module-events"},{"type":"function","title":"Agens.Job.get_config/1","doc":"Retrieves the Job configuration by Job name or `pid`.","ref":"Agens.Job.html#get_config/1"},{"type":"function","title":"Agens.Job.run/2","doc":"Runs a Job with the given input by Job name or `pid`.\n\nA supervised process for the Job must be started first using `start/1`.","ref":"Agens.Job.html#run/2"},{"type":"function","title":"Agens.Job.start/1","doc":"Starts a new Job process using the provided `Agens.Job.Config`.\n\n`start/1` does not run the Job, only starts the supervised process. See `run/2` for running the Job.","ref":"Agens.Job.html#start/1"},{"type":"module","title":"Agens.Job.Config","doc":"The Config struct defines the details of a Job.","ref":"Agens.Job.Config.html"},{"type":"module","title":"Fields - Agens.Job.Config","doc":"- `name` - An atom that identifies the Job.\n- `description` - An optional string to be added to the LM prompt that describes the basic goal of the Job.\n- `steps` - A list of `Agens.Job.Step` structs that define the sequence of agent actions to be performed.","ref":"Agens.Job.Config.html#module-fields"},{"type":"type","title":"Agens.Job.Config.t/0","doc":"","ref":"Agens.Job.Config.html#t:t/0"},{"type":"module","title":"Agens.Job.Step","doc":"The Step struct defines a single step within a Job.","ref":"Agens.Job.Step.html"},{"type":"module","title":"Fields - Agens.Job.Step","doc":"- `agent` - The name of the agent to be used in the Step.\n- `objective` - An optional string to be added to the LM prompt explaining the purpose of the Step.\n- `conditions` - An optional conditions map to control flow based on the result of the agent.","ref":"Agens.Job.Step.html#module-fields"},{"type":"type","title":"Agens.Job.Step.t/0","doc":"","ref":"Agens.Job.Step.html#t:t/0"},{"type":"module","title":"Agens.Serving","doc":"The Serving module provides functions for starting, stopping and running Servings.\n\n`Agens.Serving` accepts a `GenServer` module or `Nx.Serving` struct for processing messages.\n\n`Agens.Serving` is decoupled from `Agens.Agent` in order to reuse a single LM across multiple agents. In most cases, however, you will only need to start one text generation serving to be used by most, if not all, agents.\n\nIn some cases, you may have additional servings for more specific use cases such as image generation, speech recognition, etc.\n\nServings were built with the `Bumblebee` library in mind, as well as `Nx.Serving`. `GenServer` is supported for working with LM APIs instead, which may be more cost effective and easier to get started with.","ref":"Agens.Serving.html"},{"type":"function","title":"Agens.Serving.run/1","doc":"Executes an `Agens.Message` against an `Agens.Serving`","ref":"Agens.Serving.html#run/1"},{"type":"function","title":"Agens.Serving.start/1","doc":"Starts an `Agens.Serving` process","ref":"Agens.Serving.html#start/1"},{"type":"function","title":"Agens.Serving.stop/1","doc":"Stops an `Agens.Serving` process","ref":"Agens.Serving.html#stop/1"},{"type":"module","title":"Agens.Serving.Config","doc":"The Config struct represents the configuration for a Serving process.","ref":"Agens.Serving.Config.html"},{"type":"module","title":"Fields - Agens.Serving.Config","doc":"- `:name` - The name of the `Agens.Serving` process.\n- `:serving` - The `Nx.Serving` struct or `GenServer` module for the `Agens.Serving`.","ref":"Agens.Serving.Config.html#module-fields"},{"type":"type","title":"Agens.Serving.Config.t/0","doc":"","ref":"Agens.Serving.Config.html#t:t/0"},{"type":"behaviour","title":"Agens.Tool","doc":"The Tool behaviour.\n\nA Tool is a module that implements the `Agens.Tool` behaviour. It is used to define\nthe functionality of a tool that can be used by an `Agens.Agent`.\n\nA Tool defines the following callbacks:\n\n  - `pre/1` - pre-processes the input from the previous step before it is added to the LM prompt.\n  - `instructions/0` - returns the tool-specific instructions for the LM.\n  - `to_args/1` - parses the LM result into arguments to be used by `execute/1`.\n  - `execute/1` - executes the tool with the given arguments.\n  - `post/1` - handles the various outputs of `execute/1`, whether a map or error tuple, and returns a string for the next Step of the Job.","ref":"Agens.Tool.html"},{"type":"callback","title":"Agens.Tool.execute/1","doc":"Execute the tool with the given arguments. LM is responsible for generating arguments for the Tool based on Tool instructions, input from previous Step, Agent config and any additional context.","ref":"Agens.Tool.html#c:execute/1"},{"type":"callback","title":"Agens.Tool.instructions/0","doc":"Instructions for the LM, combined with the input from previous Step (and optionally pre-processed using `pre/1`) as well as the prompt from the Agent config.","ref":"Agens.Tool.html#c:instructions/0"},{"type":"callback","title":"Agens.Tool.post/1","doc":"Handles the various outputs of `execute/1`, whether a map or error tuple, and returns a string for the next Step of the Job.","ref":"Agens.Tool.html#c:post/1"},{"type":"callback","title":"Agens.Tool.pre/1","doc":"Pre-process the input from previous Step before it is added to the LM prompt.","ref":"Agens.Tool.html#c:pre/1"},{"type":"callback","title":"Agens.Tool.to_args/1","doc":"Parse the LM result into arguments to be used by `execute/1`.","ref":"Agens.Tool.html#c:to_args/1"},{"type":"extras","title":"Agens","doc":"![](https://github.com/jessedrelick/agens/actions/workflows/main.yml/badge.svg)\n[![Hexdocs](https://img.shields.io/badge/hex-docs-blue.svg)](https://hexdocs.pm/agens)\n[![Hex.pm](https://img.shields.io/hexpm/v/agens.svg)](https://hex.pm/packages/agens)\n[![codecov](https://codecov.io/gh/jessedrelick/agens/graph/badge.svg?token=KTJXB4SGCJ)](https://codecov.io/gh/jessedrelick/agens)\n\n__Agens__ is an Elixir application designed to build multi-agent workflows with language models.\n\nDrawing inspiration from popular tools in the Python ecosystem, such as [LangChain](https://www.langchain.com/)/[LangGraph](https://www.langchain.com/langgraph) and [CrewAI](https://www.crewai.com/), __Agens__ showcases Elixir’s unique strengths in multi-agent workflows. While the ML/AI landscape is dominated by Python, Elixir’s use of the BEAM virtual machine and OTP (Open Telecom Platform), specifically GenServers and Supervisors, makes it particularly well-suited for these tasks. Agens aims to demonstrate how these inherent design features can be leveraged effectively.\n\nBy combining Agens with powerful Elixir libraries like [Bumblebee](https://github.com/elixir-nx/bumblebee) and [Nx.Serving](https://hexdocs.pm/nx/Nx.Serving.html), along with [structured outputs in the OpenAI API](https://openai.com/index/introducing-structured-outputs-in-the-api/) and the continuous improvement of open-source language models, the reliance on Python for multi-agent workflows is significantly reduced. This shift allows Elixir’s concurrency model to truly shine.\n\n> **⚠️ Experimental:** v0.1  \n>  \n> Agens is currently an experimental project. As of version 0.1, it is primarily a proof-of-concept and learning tool. \n>  \n> The next phase of the project focuses on developing real-world examples to uncover potential issues or gaps that the current test suite may not address.  \n>  \n> These examples are designed to not only help you get started but also to advance Agens towards becoming a production-ready tool, suitable for integration into new or existing Elixir applications.","ref":"readme.html"},{"type":"extras","title":"Installation - Agens","doc":"Add `agens` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:agens, \"~> 0.1.1\"}\n  ]\nend\n```","ref":"readme.html#installation"},{"type":"extras","title":"Configuration - Agens","doc":"Future versions of Agens will be configurable by providing options to `Agens.Supervisor` in order to [avoid using application configuration](https://hexdocs.pm/elixir/1.17.2/design-anti-patterns.html#using-application-configuration-for-libraries). For now, however, you can change the Agens `Registry` if needed via config:\n\n```elixir\nconfig :agens, registry: Agens.Registry\n```\n\nIn addition, you can also override the prompt prefixes:\n\n```elixir\nconfig :agens, prompts: %{\n  prompt:\n    {\"Agent\",\n      \"You are a specialized agent with the following capabilities and expertise\"},\n  identity:\n    {\"Identity\",\n      \"You are a specialized agent with the following capabilities and expertise\"},\n  context: {\"Context\", \"The purpose or goal behind your tasks are to\"},\n  constraints:\n    {\"Constraints\", \"You must operate with the following constraints or limitations\"},\n  examples:\n    {\"Examples\",\n      \"You should consider the following examples before returning results\"},\n  reflection:\n    {\"Reflection\",\n      \"You should reflect on the following factors before returning results\"},\n  instructions:\n    {\"Tool Instructions\",\n      \"You should provide structured output for function calling based on the following instructions\"},\n  objective: {\"Step Objective\", \"The objective of this step is to\"},\n  description:\n    {\"Job Description\", \"This is part of multi-step job to achieve the following\"},\n  input:\n    {\"Input\",\n      \"The following is the actual input from the user, system or another agent\"}\n}\n```\n\nSee the [Prompting](#prompting) section below or `Agens.Message` for more information on prompt prefixes.","ref":"readme.html#configuration"},{"type":"extras","title":"Usage - Agens","doc":"Building a multi-agent workflow with Agens involves a few different steps and core entities:\n\n---\n**1. Add the Agens Supervisor to your Supervision tree**\n\nThis will start Agens as a supervised process inside your application:\n\n```elixir\nSupervisor.start_link(\n  [\n    {Agens.Supervisor, name: Agens.Supervisor}\n  ],\n  strategy: :one_for_one\n)\n```\n\nSee `Agens.Supervisor` for more information\n\n---\n**2. Start one or more Servings**\n\nA **Serving** is essentially a wrapper for language model inference. It can be an `Nx.Serving` struct, either returned by `Bumblebee` or manually created, or a `GenServer` that interfaces with the OpenAI API or other language model APIs. Technically, due to `GenServer` support, a Serving doesn't need to be limited to language models or machine learning—it can also handle regular API calls.\n\n```elixir\nApplication.put_env(:nx, :default_backend, EXLA.Backend)\nauth_token = System.get_env(\"HF_AUTH_TOKEN\")\n\nmy_serving = fn ->\n  repo = {:hf, \"mistralai/Mistral-7B-Instruct-v0.2\", auth_token: auth_token}\n\n  {:ok, model} = Bumblebee.load_model(repo, type: :bf16)\n  {:ok, tokenizer} = Bumblebee.load_tokenizer(repo)\n  {:ok, generation_config} = Bumblebee.load_generation_config(repo)\n\n  Bumblebee.Text.generation(model, tokenizer, generation_config)\nend\n\nserving_config = %Agens.Serving.Config{\n  name: :my_serving,\n  serving: my_serving()\n}\n\n{:ok, pid} = Agens.Serving.start(serving_config)\n```\n\nSee `Agens.Serving` for more information\n\n---\n**3. Create and start one or more Agents**\n\nAn **Agent** in the context of Agens is responsible for communicating with Servings and can provide additional context during these interactions.\n\nIn practice, Agents typically have their own specialized tasks or capabilities while communicating with the same Serving. Many projects may use a single Serving, such as a language model (LM) or an LM API, but employ multiple Agents to perform different tasks using that Serving. \n\nAdditionally, Agents can use modules implementing the `Agens.Tool` behaviour to extend their capabilities beyond standard LM inference, enabling function-calling and other advanced operations.\n\n```elixir\nagent_config = %Agens.Agent.Config{\n  name: :my_agent,\n  serving: :my_serving\n}\n{:ok, pid} = Agens.Agent.start(agent_config)\n```\n\nSee `Agens.Agent` for more information\n\n---\n**4. Create and start one or more Jobs**\n\nWhile Agens is designed to be flexible enough to allow direct communication with an `Agens.Serving` or `Agens.Agent`, its primary goal is to facilitate a multi-agent workflow that uses various steps to achieve a final result. Each step (`Agens.Job.Step`) employs an Agent to accomplish its objective, and the results are then passed to the next step in the **Job**. Conditions can also be used to determine the routing between steps or to conclude the job.\n\n```elixir\njob_config = %Agens.Job.Config{\n  name: :my_job,\n  description: \"an example job\",\n  steps: [\n    %Agens.Job.Step{\n      agent: :my_agent,\n      objective: \"first step objective\"\n    },\n    %Agens.Job.Step{\n      agent: :my_agent,\n      conditions: %{\n        \"__DEFAULT__\" => :end\n      }\n    }\n  ]\n}\n{:ok, pid} = Agens.Job.start(job_config)\nAgens.Job.run(:my_job, \"user input\")\n```\n\nSee `Agens.Job` for more information\n\n---","ref":"readme.html#usage"},{"type":"extras","title":"Prompting - Agens","doc":"Agens provides a variety of different ways to customize the final prompt sent to the language model (LM) or Serving. A natural language string can be assigned to the entity's specialized field (see below), while `nil` values will omit that field from the final prompt. This approach allows for precise control over the prompt’s content.\n\nAll fields with values, in addition to user input, will be included in the final prompt !!!!using the [in-context learning]() method!!!!. The goal should be to balance detailed prompts with efficient token usage by focusing on relevant fields and using concise language. This approach will yield the best results with minimal token usage, keeping costs low and performance high.","ref":"readme.html#prompting"},{"type":"extras","title":"User/Agent - Agens","doc":"The `input` value is the only required field for building prompts. This value can be the initial value provided to `Agens.Job.run/2`, or the final result of a previous step (`Agens.Job.Step`). Both the `input` and `result` are stored in `Agens.Message`, which can also be used to send messages directly to `Agens.Agent` or `Agens.Serving` without being part of an `Agens.Job`.","ref":"readme.html#user-agent"},{"type":"extras","title":"Job - Agens","doc":"`Agens.Job.Config` uses the `description` field to configure the prompt for all messages within the Job. This field should be used carefully as it will be sent to the Serving with every prompt.","ref":"readme.html#job"},{"type":"extras","title":"Step - Agens","doc":"`Agens.Job.Step` uses the `objective` field to customize the final prompt sent to the Serving. This can provide more specific information in the final prompt than the Job `description` or Agent `prompt`.","ref":"readme.html#step"},{"type":"extras","title":"Agent - Agens","doc":"`Agens.Agent` provides the most advanced prompt capabilities. The `prompt` field of `Agens.Agent.Config` accepts either a simple string value, or an `Agens.Agent.Prompt` struct. The following fields, which are all optional, can be used with the struct approach:\n\n- `:identity` - a string representing the purpose and capabilities of the agent\n- `:context` - a string representing the goal or purpose of the agent's actions\n- `:constraints` - a string listing any constraints or limitations on the agent's actions\n- `:examples` - a list of maps representing example inputs and outputs for the agent\n- `:reflection` - a string representing any additional considerations or reflection the agent should make before returning results\n\nKeep in mind that a single agent can be used across multiple jobs, so it is best to restrict the agent prompt to specific capabilities and use `objective` on `Agens.Job.Step` or `description` on `Agens.Job.Config` for Job or Step-specific prompting.","ref":"readme.html#agent"},{"type":"extras","title":"Tool - Agens","doc":"When creating Tools with the `Agens.Tool` behaviour, the `c:Agens.Tool.instructions/0` callback can be used to include specific instructions in the final prompt. These instructions may also include examples, especially for structured output, which can be crucial for designing a Tool that delivers predictable results.\n\nIt is important to note that these instructions are provided to the Serving **before** the Tool is used, ensuring that the language model (LM) supplies the correct inputs to the Tool. After receiving these inputs, the Tool should be able to generate the relevant arguments to make the function call, and finally provide the expected output for the next step of the job.\n\nSee `Agens.Tool` for more information on using Tools.","ref":"readme.html#tool"},{"type":"extras","title":"Summary - Agens","doc":"- **User/Agent**: `input`/`result`\n- **Job**: `description`\n- **Agent**: `prompt` (`string` or `Agens.Agent.Prompt`)\n- **Step**: `objective`\n- **Tool**: `instructions`\n\n> **Note:** \n>\n> Depending on your use case, some fields may be more relevant than others.\n>\n> It’s often beneficial to be more descriptive at granular levels, such as the `objective` of `Agens.Job.Step` or the `instructions` for `Agens.Tool`, while taking a more minimal approach with higher-level fields, such as the `description` of `Agens.Job.Config` or the `prompt` of `Agens.Agent.Config`.","ref":"readme.html#summary"},{"type":"extras","title":"Name - Agens","doc":"The name Agens comes from the Latin word for 'Agents' or 'Actors.' It also draws from **intellectus agens**, a term in medieval philosophy meaning ['active intellect'](https://en.wikipedia.org/wiki/Active_intellect), which describes the mind’s ability to actively process and abstract information. This reflects the goal of the Agens project: to create intelligent, autonomous agents that manage workflows within the Elixir ecosystem.","ref":"readme.html#name"},{"type":"extras","title":"License - Agens","doc":"This project is licensed under the Apache License, Version 2.0. See the [LICENSE](./LICENSE) file for more details.","ref":"readme.html#license"},{"type":"extras","title":"LICENSE","doc":"Apache License\n                           Version 2.0, January 2004\n                        http://www.apache.org/licenses/\n\n   TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION\n\n   1. Definitions.\n\n      \"License\" shall mean the terms and conditions for use, reproduction,\n      and distribution as defined by Sections 1 through 9 of this document.\n\n      \"Licensor\" shall mean the copyright owner or entity authorized by\n      the copyright owner that is granting the License.\n\n      \"Legal Entity\" shall mean the union of the acting entity and all\n      other entities that control, are controlled by, or are under common\n      control with that entity. For the purposes of this definition,\n      \"control\" means (i) the power, direct or indirect, to cause the\n      direction or management of such entity, whether by contract or\n      otherwise, or (ii) ownership of fifty percent (50%) or more of the\n      outstanding shares, or (iii) beneficial ownership of such entity.\n\n      \"You\" (or \"Your\") shall mean an individual or Legal Entity\n      exercising permissions granted by this License.\n\n      \"Source\" form shall mean the preferred form for making modifications,\n      including but not limited to software source code, documentation\n      source, and configuration files.\n\n      \"Object\" form shall mean any form resulting from mechanical\n      transformation or translation of a Source form, including but\n      not limited to compiled object code, generated documentation,\n      and conversions to other media types.\n\n      \"Work\" shall mean the work of authorship, whether in Source or\n      Object form, made available under the License, as indicated by a\n      copyright notice that is included in or attached to the work\n      (an example is provided in the Appendix below).\n\n      \"Derivative Works\" shall mean any work, whether in Source or Object\n      form, that is based on (or derived from) the Work and for which the\n      editorial revisions, annotations, elaborations, or other modifications\n      represent, as a whole, an original work of authorship. For the purposes\n      of this License, Derivative Works shall not include works that remain\n      separable from, or merely link (or bind by name) to the interfaces of,\n      the Work and Derivative Works thereof.\n\n      \"Contribution\" shall mean any work of authorship, including\n      the original version of the Work and any modifications or additions\n      to that Work or Derivative Works thereof, that is intentionally\n      submitted to Licensor for inclusion in the Work by the copyright owner\n      or by an individual or Legal Entity authorized to submit on behalf of\n      the copyright owner. For the purposes of this definition, \"submitted\"\n      means any form of electronic, verbal, or written communication sent\n      to the Licensor or its representatives, including but not limited to\n      communication on electronic mailing lists, source code control systems,\n      and issue tracking systems that are managed by, or on behalf of, the\n      Licensor for the purpose of discussing and improving the Work, but\n      excluding communication that is conspicuously marked or otherwise\n      designated in writing by the copyright owner as \"Not a Contribution.\"\n\n      \"Contributor\" shall mean Licensor and any individual or Legal Entity\n      on behalf of whom a Contribution has been received by Licensor and\n      subsequently incorporated within the Work.\n\n   2. Grant of Copyright License. Subject to the terms and conditions of\n      this License, each Contributor hereby grants to You a perpetual,\n      worldwide, non-exclusive, no-charge, royalty-free, irrevocable\n      copyright license to reproduce, prepare Derivative Works of,\n      publicly display, publicly perform, sublicense, and distribute the\n      Work and such Derivative Works in Source or Object form.\n\n   3. Grant of Patent License. Subject to the terms and conditions of\n      this License, each Contributor hereby grants to You a perpetual,\n      worldwide, non-exclusive, no-charge, royalty-free, irrevocable\n      (except as stated in this section) patent license to make, have made,\n      use, offer to sell, sell, import, and otherwise transfer the Work,\n      where such license applies only to those patent claims licensable\n      by such Contributor that are necessarily infringed by their\n      Contribution(s) alone or by combination of their Contribution(s)\n      with the Work to which such Contribution(s) was submitted. If You\n      institute patent litigation against any entity (including a\n      cross-claim or counterclaim in a lawsuit) alleging that the Work\n      or a Contribution incorporated within the Work constitutes direct\n      or contributory patent infringement, then any patent licenses\n      granted to You under this License for that Work shall terminate\n      as of the date such litigation is filed.\n\n   4. Redistribution. You may reproduce and distribute copies of the\n      Work or Derivative Works thereof in any medium, with or without\n      modifications, and in Source or Object form, provided that You\n      meet the following conditions:\n\n      (a) You must give any other recipients of the Work or\n          Derivative Works a copy of this License; and\n\n      (b) You must cause any modified files to carry prominent notices\n          stating that You changed the files; and\n\n      (c) You must retain, in the Source form of any Derivative Works\n          that You distribute, all copyright, patent, trademark, and\n          attribution notices from the Source form of the Work,\n          excluding those notices that do not pertain to any part of\n          the Derivative Works; and\n\n      (d) If the Work includes a \"NOTICE\" text file as part of its\n          distribution, then any Derivative Works that You distribute must\n          include a readable copy of the attribution notices contained\n          within such NOTICE file, excluding those notices that do not\n          pertain to any part of the Derivative Works, in at least one\n          of the following places: within a NOTICE text file distributed\n          as part of the Derivative Works; within the Source form or\n          documentation, if provided along with the Derivative Works; or,\n          within a display generated by the Derivative Works, if and\n          wherever such third-party notices normally appear. The contents\n          of the NOTICE file are for informational purposes only and\n          do not modify the License. You may add Your own attribution\n          notices within Derivative Works that You distribute, alongside\n          or as an addendum to the NOTICE text from the Work, provided\n          that such additional attribution notices cannot be construed\n          as modifying the License.\n\n      You may add Your own copyright statement to Your modifications and\n      may provide additional or different license terms and conditions\n      for use, reproduction, or distribution of Your modifications, or\n      for any such Derivative Works as a whole, provided Your use,\n      reproduction, and distribution of the Work otherwise complies with\n      the conditions stated in this License.\n\n   5. Submission of Contributions. Unless You explicitly state otherwise,\n      any Contribution intentionally submitted for inclusion in the Work\n      by You to the Licensor shall be under the terms and conditions of\n      this License, without any additional terms or conditions.\n      Notwithstanding the above, nothing herein shall supersede or modify\n      the terms of any separate license agreement you may have executed\n      with Licensor regarding such Contributions.\n\n   6. Trademarks. This License does not grant permission to use the trade\n      names, trademarks, service marks, or product names of the Licensor,\n      except as required for reasonable and customary use in describing the\n      origin of the Work and reproducing the content of the NOTICE file.\n\n   7. Disclaimer of Warranty. Unless required by applicable law or\n      agreed to in writing, Licensor provides the Work (and each\n      Contributor provides its Contributions) on an \"AS IS\" BASIS,\n      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or\n      implied, including, without limitation, any warranties or conditions\n      of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A\n      PARTICULAR PURPOSE. You are solely responsible for determining the\n      appropriateness of using or redistributing the Work and assume any\n      risks associated with Your exercise of permissions under this License.\n\n   8. Limitation of Liability. In no event and under no legal theory,\n      whether in tort (including negligence), contract, or otherwise,\n      unless required by applicable law (such as deliberate and grossly\n      negligent acts) or agreed to in writing, shall any Contributor be\n      liable to You for damages, including any direct, indirect, special,\n      incidental, or consequential damages of any character arising as a\n      result of this License or out of the use or inability to use the\n      Work (including but not limited to damages for loss of goodwill,\n      work stoppage, computer failure or malfunction, or any and all\n      other commercial damages or losses), even if such Contributor\n      has been advised of the possibility of such damages.\n\n   9. Accepting Warranty or Additional Liability. While redistributing\n      the Work or Derivative Works thereof, You may choose to offer,\n      and charge a fee for, acceptance of support, warranty, indemnity,\n      or other liability obligations and/or rights consistent with this\n      License. However, in accepting such obligations, You may act only\n      on Your own behalf and on Your sole responsibility, not on behalf\n      of any other Contributor, and only if You agree to indemnify,\n      defend, and hold each Contributor harmless for any liability\n      incurred by, or claims asserted against, such Contributor by reason\n      of your accepting any such warranty or additional liability.\n\n   END OF TERMS AND CONDITIONS\n\n   APPENDIX: How to apply the Apache License to your work.\n\n      To apply the Apache License to your work, attach the following\n      boilerplate notice, with the fields enclosed by brackets \"[]\"\n      replaced with your own identifying information. (Don't include\n      the brackets!)  The text should be enclosed in the appropriate\n      comment syntax for the file format. We also recommend that a\n      file or class name and description of purpose be included on the\n      same \"printed page\" as the copyright notice for easier\n      identification within third-party archives.\n\n   Copyright 2024 Jesse Drelick\n\n   Licensed under the Apache License, Version 2.0 (the \"License\");\n   you may not use this file except in compliance with the License.\n   You may obtain a copy of the License at\n\n       http://www.apache.org/licenses/LICENSE-2.0\n\n   Unless required by applicable law or agreed to in writing, software\n   distributed under the License is distributed on an \"AS IS\" BASIS,\n   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.\n   See the License for the specific language governing permissions and\n   limitations under the License.","ref":"license.html"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,52,46,50]}}